"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const debug_1 = __importDefault(require("debug"));
const zen_observable_1 = __importDefault(require("zen-observable"));
const observable_promise_1 = require("../observable-promise");
const implementation_1 = __importDefault(require("./implementation"));
const thread_1 = require("./thread");
exports.Thread = thread_1.Thread;
let nextPoolID = 1;
const hasSymbols = () => typeof Symbol === 'function';
const hasSymbol = (name) => hasSymbols() && Boolean(Symbol[name]);
function flatMap(array, mapper) {
    return array.reduce((flattened, element) => [...flattened, ...mapper(element)], []);
}
function sleep(ms) {
    return new Promise(resolve => setTimeout(resolve, ms));
}
function slugify(text) {
    return text.replace(/\W/g, " ").trim().replace(/\s+/g, "-");
}
var PoolEventType;
(function (PoolEventType) {
    PoolEventType["initialized"] = "initialized";
    PoolEventType["taskCompleted"] = "taskCompleted";
    PoolEventType["taskFailed"] = "taskFailed";
    PoolEventType["taskQueued"] = "taskQueued";
    PoolEventType["taskQueueDrained"] = "taskQueueDrained";
    PoolEventType["taskStart"] = "taskStart";
    PoolEventType["terminated"] = "terminated";
})(PoolEventType = exports.PoolEventType || (exports.PoolEventType = {}));
function createArray(size) {
    const array = [];
    for (let index = 0; index < size; index++) {
        array.push(index);
    }
    return array;
}
function findIdlingWorker(workers, maxConcurrency) {
    return workers.find(worker => worker.runningJobs.length < maxConcurrency);
}
function spawnWorkers(spawnWorker, count) {
    return createArray(count).map(() => ({
        init: spawnWorker(),
        runningJobs: []
    }));
}
function PoolConstructor(spawnWorker, optionsOrSize) {
    const options = typeof optionsOrSize === "number"
        ? { size: optionsOrSize }
        : optionsOrSize || {};
    const debug = debug_1.default(`threads:pool:${slugify(options.name || String(nextPoolID++))}`);
    const { concurrency = 1, size = implementation_1.default.defaultPoolSize } = options;
    let isClosing = false;
    let nextTaskID = 1;
    const taskQueue = [];
    const workers = spawnWorkers(spawnWorker, size);
    let eventSubject;
    const eventObservable = observable_promise_1.makeHot(new zen_observable_1.default(subscriber => {
        eventSubject = subscriber;
    }));
    Promise.all(workers.map(worker => worker.init)).then(() => eventSubject.next({
        type: PoolEventType.initialized,
        size: workers.length
    }), error => eventSubject.error(error));
    const scheduleWork = () => {
        debug(`Attempt de-queueing a task in order to run it...`);
        const availableWorker = findIdlingWorker(workers, concurrency);
        if (!availableWorker)
            return;
        const nextTask = taskQueue.shift();
        if (!nextTask) {
            debug(`Task queue is empty`);
            eventSubject.next({ type: PoolEventType.taskQueueDrained });
            return;
        }
        const workerID = workers.indexOf(availableWorker) + 1;
        debug(`Running task #${nextTask.id} on worker #${workerID}...`);
        eventSubject.next({
            type: PoolEventType.taskStart,
            taskID: nextTask.id,
            workerID
        });
        const run = (worker, task) => __awaiter(this, void 0, void 0, function* () {
            const removeJobFromWorkersRunningJobs = () => {
                worker.runningJobs = worker.runningJobs.filter(someRunPromise => someRunPromise !== runPromise);
            };
            // Defer job execution by one tick to give handlers time to subscribe
            yield sleep(0);
            try {
                const returnValue = yield task.run(yield availableWorker.init);
                debug(`Task #${nextTask.id} completed successfully`);
                removeJobFromWorkersRunningJobs();
                eventSubject.next({
                    type: PoolEventType.taskCompleted,
                    returnValue,
                    taskID: nextTask.id,
                    workerID
                });
            }
            catch (error) {
                debug(`Task #${nextTask.id} failed`);
                removeJobFromWorkersRunningJobs();
                eventSubject.next({
                    type: PoolEventType.taskFailed,
                    taskID: nextTask.id,
                    error,
                    workerID
                });
            }
            finally {
                if (!isClosing) {
                    scheduleWork();
                }
            }
        });
        const runPromise = run(availableWorker, nextTask);
        availableWorker.runningJobs.push(runPromise);
    };
    const pool = {
        completed(allowResolvingImmediately = false) {
            return __awaiter(this, void 0, void 0, function* () {
                const getCurrentlyRunningJobs = () => flatMap(workers, worker => worker.runningJobs);
                if (allowResolvingImmediately && taskQueue.length === 0) {
                    return Promise.all(getCurrentlyRunningJobs());
                }
                const poolEventPromise = new Promise((resolve, reject) => {
                    const subscription = eventObservable.subscribe(event => {
                        if (event.type === PoolEventType.taskQueueDrained) {
                            subscription.unsubscribe();
                            resolve();
                        }
                        else if (event.type === PoolEventType.taskFailed) {
                            subscription.unsubscribe();
                            reject(event.error);
                        }
                    });
                });
                yield Promise.race([
                    poolEventPromise,
                    eventObservable // make a pool-wide error reject the completed() result promise
                ]);
                yield Promise.all(getCurrentlyRunningJobs());
            });
        },
        events() {
            return eventObservable;
        },
        queue(taskFunction) {
            if (isClosing) {
                throw Error(`Cannot schedule pool tasks after terminate() has been called.`);
            }
            const task = {
                id: nextTaskID++,
                run: taskFunction
            };
            debug(`Queueing task #${task.id}...`);
            taskQueue.push(task);
            eventSubject.next({
                type: PoolEventType.taskQueued,
                taskID: task.id
            });
            const resultPromise = new Promise((resolve, reject) => {
                const eventSubscription = pool.events().subscribe(event => {
                    if (event.type === PoolEventType.taskCompleted && event.taskID === task.id) {
                        eventSubscription.unsubscribe();
                        resolve(event.returnValue);
                    }
                    else if (event.type === PoolEventType.taskFailed && event.taskID === task.id) {
                        eventSubscription.unsubscribe();
                        reject(event.error);
                    }
                    else if (event.type === PoolEventType.terminated) {
                        eventSubscription.unsubscribe();
                        reject(Error("Pool has been terminated before task was run."));
                    }
                });
                try {
                    scheduleWork();
                }
                catch (error) {
                    eventSubscription.unsubscribe();
                    reject(error);
                }
            });
            // Don't raise an UnhandledPromiseRejection error if not handled
            // Reason: Because we just return this promise for convenience, but usually only
            //         pool.completed() will be used, leaving this quasi-duplicate promise unhandled.
            resultPromise.catch(() => undefined);
            return resultPromise;
        },
        terminate(force) {
            return __awaiter(this, void 0, void 0, function* () {
                isClosing = true;
                if (!force) {
                    yield pool.completed(true);
                }
                eventSubject.next({
                    type: PoolEventType.terminated,
                    remainingQueue: [...taskQueue]
                });
                eventSubject.complete();
                yield Promise.all(workers.map((worker) => __awaiter(this, void 0, void 0, function* () { return thread_1.Thread.terminate(yield worker.init); })));
            });
        }
    };
    if (hasSymbols() && hasSymbol("toStringTag")) {
        pool[Symbol.toStringTag] = () => `[object Pool]`;
    }
    return pool;
}
PoolConstructor.EventType = PoolEventType;
exports.Pool = PoolConstructor;
